#!/bin/sh

prefix=/opt/dreamcast
shtarget=sh-sega-dreamcast
armtarget=arm-elf
binutils=binutils-2.11.2
gcc=gcc-3.0.4
gcc_core=`echo $gcc|sed 's/-/-core-/'`
gcc_gpp=`echo $gcc|sed 's/-/-g++-/'`
newlib=newlib-1.9.0
gnubase=ftp://ftp.gnu.org/gnu
redhatbase=ftp://sources.redhat.com/pub
patchbase=http://our-arca.de/Download/Dreamcast

[ -d Makedevel ] || mkdir -p Makedevel
cd Makedevel

[ -d archives ] || mkdir -p archives

cd archives
[ -f ${binutils}.tar.gz ] || wget ${gnubase}/binutils/${binutils}.tar.gz
[ -f ${gcc_core}.tar.gz ] || wget ${gnubase}/gcc/${gcc}/${gcc_core}.tar.gz
[ -f ${gcc_gpp}.tar.gz ] || wget ${gnubase}/gcc/${gcc}/${gcc_gpp}.tar.gz
[ -f ${newlib}.tar.gz ] || wget ${redhatbase}/newlib/${newlib}.tar.gz
[ -f ${binutils}.patch.gz ] || wget ${patchbase}/${binutils}.patch.gz
[ -f ${gcc}.patch.gz ] || wget ${patchbase}/${gcc}.patch.gz
[ -f ${newlib}.patch.gz ] || wget {$patchbase}/${newlib}.patch.gz

cd ../

bindir=$prefix/bin
mkdir -p $prefix
PATH=$bindir:$PATH

echo "Cleaning up..."
rm -fr build-* $gcc $binutils $newlib

echo "Unpacking binutils..."
tar zxf archives/${binutils}.tar.gz
gzip -dc archives/${binutils}.patch.gz | patch -p0

echo "Unpacking gcc..."
tar zxf archives/${gcc_core}.tar.gz
tar zxf archives/${gcc_gpp}.tar.gz
gzip -dc archives/${gcc}.patch.gz | patch -p0

echo "Unpacking newlib..."
tar zxf archives/${newlib}.tar.gz
gzip -dc archives/${newlib}.patch.gz | patch -p0

set -e

echo "Building binutils..."
mkdir build-sh-binutils
cd build-sh-binutils
../$binutils/configure \
  --target=$shtarget \
  --prefix=$prefix
make all install

cd ../

echo "Building gcc..."
mkdir build-sh-gcc
cd build-sh-gcc
../$gcc/configure \
  --target=$shtarget \
  --prefix=$prefix \
  --without-headers \
  --with-newlib \
  --enable-languages=c
make all-gcc
make install-gcc

cd ../

echo "Building newlib..."
mkdir build-sh-newlib
cd build-sh-newlib
../$newlib/configure \
  --target=$shtarget \
  --prefix=$prefix
make all install \
  CC_FOR_TARGET=$bindir/${shtarget}-gcc \
  AS_FOR_TARGET=$bindir/${shtarget}-as \
  LD_FOR_TARGET=$bindir/${shtarget}-ld \
  AR_FOR_TARGET=$bindir/${shtarget}-ar \
  RANLIB_FOR_TARGET=$bindir/${shtarget}-ranlib

cd ../

echo "Building gcc,g++..."
cd build-sh-gcc
../$gcc/configure \
  --target=$shtarget \
  --prefix=$prefix \
  --without-headers \
  --with-newlib \
  --enable-languages=c,c++
make all install

(cd $shtarget/bin
for i in sh-sega-dreamcast-*
  do ln -sv $i `echo $i|sed s@sega-dreamcast@elf@`
done
)
cd ../

mkdir build-arm-binutils
cd build-arm-binutils

../$binutils/configure \
  --target=$armtarget \
  --prefix=$prefix
make all install

cd ../

(cd $prefix/$armtarget ; ln -s ../$shtarget/include)

mkdir build-arm-gcc
cd build-arm-gcc
../$gcc/configure \
  --target=$armtarget \
  --prefix=$prefix \
  --without-headers \
  --without-lib \
  --enable-languages=c
make all install

cd ../

