/*
 * This file is part of the DCStella project
 * location: dcstella/aica/main.c
 * written by SvOlli
 */

#include "aica.h"
#include "TIASound.h"

volatile unsigned long *cmdstart = (volatile unsigned long *) 0xf000;
volatile unsigned long *cmdend   = (volatile unsigned long *) 0xffff;
volatile unsigned long *cmdpos;

#define fragmentRAMbase 0xe000
#define fragmentSize    0x0400
#define fragments       3
unsigned long fragmentRAM[fragments];

/* sampleRate needs to be 31400 or a change in TIASound.c is needed */
#define sampleRate   31400

extern volatile int timer;

#if 0
void timer_wait(int jiffies)
{
  int fin = timer + jiffies;
  while (timer <= fin)
    ;
}
#endif

int arm_main()
{
  int bufferindex = 0;
  volatile unsigned long playpos;
  unsigned long i;
  int mute = 0;
  int playing = 0;
  
  /* Initialize the AICA part of the SPU */
  aica_init();
  
  /* Initialize the TIA sound emulation */
  Tia_sound_init(31400, sampleRate);
  
  /* Observe channel 0 */
  SNDREG8(0x280d) = 0;

  /* Clear the audio buffers */
  for(i = fragmentRAMbase;
      i < fragmentRAMbase+fragments*fragmentSize;
      i+=4)
    *(unsigned long*)i = 0;

  /* Set the beginning of the fragments in RAM */
  for(i = 0; i < fragments; i++)
    fragmentRAM[i] = fragmentRAMbase + fragmentSize * i;

  /* The main loop */
  for(cmdpos=cmdstart;;)
  {
    while(*cmdpos)
    {
      switch((*cmdpos) & 0xff0000e0)
      {
       case 0x01000000:    /* Set AUDC0 */
	audc_0 = (*cmdpos) & 0x0f;
	update_chan_0();
	break;
	
       case 0x01000020:    /* Set AUDC1 */
	audc_1 = (*cmdpos) & 0x0f;
	update_chan_1();
	break;
	
       case 0x01000040:    /* Set AUDF0 */
	audf_0 = (*cmdpos) & 0x1f;
	update_chan_0();
	break;
	
       case 0x01000060:    /* Set AUDF1 */
	audf_1 = (*cmdpos) & 0x1f;
	update_chan_1();
	break;
	
       case 0x01000080:    /* Set AUDV0 */
	audv_0 = ((*cmdpos) & 0x0f) << 3;
	update_chan_0();
	break;
	
       case 0x010000a0:    /* Set AUDV1 */
	audv_1 = ((*cmdpos) & 0x0f) << 3;
	update_chan_1();
	break;
	
       case 0x010000c0:    /* Quit */
	/* not implemeted (why should?) */
	break;
	
       case 0x010000e0:    /* Change mute command */
	mute = ((*cmdpos) & 0x01);
	break;
	
       default:
	break;
      }
    
      *cmdpos=0;
      cmdpos++;
      if(cmdpos >= cmdend) cmdpos=cmdstart;
    }
    
    /* Choose new buffer */
    bufferindex++;
    if( bufferindex >= fragments ) bufferindex = 0;
    
    if(mute)
    {
      /* Stop playing */
      playing = 0;
      aica_stop(0);
    }
    else
    {
      if( !playing )
      {
	playing = 1;
	aica_play(0,fragmentRAMbase, SM_16BIT, 0, 
		  fragmentSize/2 * fragments - 1,
		  sampleRate, 240, 128, 1/*repeat*/);
      }
      
      /* what as long as the playing cursor is in the same
       * fragment as the calulating cursor */
      playpos = SNDREG32(0x2814);
      while( bufferindex == ( playpos / (fragmentSize/2) ) )
      {
	playpos = SNDREG32(0x2814);
      }
      

      /* Get new data */
      Tia_process(fragmentRAM[bufferindex], fragmentSize);
    }
    
    /* All done, begin again */
  }
}
