/*
 * This file is part of the DCStella project
 * location: dcstella/core/mainDC.cxx
 * written by SvOlli, based on the original Stella mainX11.cxx
 */

#define DCSTELLA "DCStella"
#define VERSION  "V0.2f"
#define CONFIG_FILE "/dcstella.ini"


#include <assert.h>
#include <fstream.h>
#include <iostream.h>
#include <strstream.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include <string>

#include "bspf.hxx"
#include "Console.hxx"
#include "Event.hxx"
#include "MediaSrc.hxx"
#include "PropsSet.hxx"
#include "sndDC.hxx"
#include "System.hxx"
#include "MD5.hxx"


#define TRUE 1
#define FALSE 0


extern "C"
{
#include <dream.h>
#include "screen.h"
#include "discio.h"
#include "vmu.h"
#include "o/compilemode.h"
}


#if (DEBUG == FALSE)
#define serial_init(x) /* nothing, since only debug version should output */
#endif


#define CHARSET_8X8_FILE        config_value_by_key("charset.8x8.file")
#define CHARSET_8X16_FILE       config_value_by_key("charset.8x16.file")
#define TITLE_IMAGE_FILE        config_value_by_key("title.image.file")
#define TITLE_IMAGE_XPOS   atoi(config_value_by_key("title.image.xpos"))
#define TITLE_IMAGE_YPOS   atoi(config_value_by_key("title.image.ypos"))
#define SELECT_IMAGE_FILE       config_value_by_key("select.image.file")
#define SELECT_IMAGE_XPOS  atoi(config_value_by_key("select.image.xpos"))
#define SELECT_IMAGE_YPOS  atoi(config_value_by_key("select.image.ypos"))
#define SELECT_MENU_XPOS   atoi(config_value_by_key("select.menu.xpos"))
#define SELECT_MENU_YPOS   atoi(config_value_by_key("select.menu.ypos"))
#define SELECT_MENU_WIDTH  atoi(config_value_by_key("select.menu.width"))
#define SELECT_MENU_HEIGHT atoi(config_value_by_key("select.menu.height"))
#define SELECT_MENUTEXT_FGCOL   config_value_by_key("select.menutext.fgcolor")
#define SELECT_MENUTEXT_BGCOL   config_value_by_key("select.menutext.bgcolor")
#define SELECT_MENUBAR_FGCOL    config_value_by_key("select.menubar.fgcolor")
#define SELECT_MENUBAR_BGCOL    config_value_by_key("select.menubar.bgcolor")
#define SELECT_AUDIO_PLAYER     config_value_by_key("select.audio.player")
#define SELECT_AUDIO_START atoi(config_value_by_key("select.audio.start"))
#define SELECT_AUDIO_FILE       config_value_by_key("select.audio.file")
#define ROMS_DIR                config_value_by_key("roms.dir")
#define ROMS_LIST_FILE          config_value_by_key("roms.list.file")
#define PROPERTIES_DIR          config_value_by_key("properties.dir")
#define PROFILES_DIR            config_value_by_key("profiles.dir")
#define AUTORUN_FILE            config_value_by_key("autorun.file")
#define CHEAT_ALLOWED           config_value_by_key("cheat.allowed")

#define MAXROMSIZE 512*1024

#define JOYMIN  80
#define JOYMAX (256-JOYMIN)

int  actualgame,topoffsetgame;

#define CONTROLLER_NONE     0
#define CONTROLLER_JOYSTICK 1
#define CONTROLLER_PADDLES  2
#define CONTROLLER_BOOSTER  3
#define CONTROLLER_KEYPAD   4
#define CONTROLLER_DRIVING  5
#define CONTROLLER_LIGHTGUN 6
#define CONTROLLER_MINDLINK 7

// TODO1.2
//PropertiesSet propertiesSet("Cartridge.Name");
PropertiesSet propertiesSet;

int emuVidMode;    // actual DC video mode
int DC_screen_size_X;     // x size of actual DC mode
int DC_screen_size_Y;     // y size of actual DC mode
uInt32 topLeft; // topLeft corner of Game to be displayed

uInt32 theColorTable[256]; // maybe uInt16

// Indicates the width and height of the game display based on properties
uInt32 theHeight;
uInt32 theWidth;
uInt32 theScanlines;
uInt32 theXOffset;
uInt32 theYOffset;

// Pointer to the console object or the null pointer
Console* theConsole;

// Event object to use
Event theEvent;

// Indicates if the entire frame should be redrawn
bool theRedrawEntireFrameFlag = true;

// Indicates if the user wants to quit
bool theQuitIndicator = false;

bool portSwapAllowed;
bool portsSwaped;

bool   cheatAllowed = false;
uInt8  cheatSavedMem[0x80];
uInt16 cheatAddress;
uInt8  cheatValue;
int    cheatCursorPos = 0;

/*
 * TODO:
 * - loadable vmu images
 */

void handleJoystick(cont_cond_t cond,int port)
{
  /*
   * handle the atari joystick controller
   */

  Int32 value;

  // LEFT
  value=
    ((cond.joyx < JOYMIN) || ((~cond.buttons) & CONT_DPAD_LEFT))
    ? 1 : 0;
  if(port == 0)
  {
    theEvent.set(Event::JoystickZeroLeft,value);
  }
  else
  {
    theEvent.set(Event::JoystickOneLeft,value);
  }
  
  // RIGHT
  value=
    ((cond.joyx > JOYMAX) || ((~cond.buttons) & CONT_DPAD_RIGHT))
    ? 1 : 0;
  if(port == 0)
  {
    theEvent.set(Event::JoystickZeroRight,value);
  }
  else
  {
    theEvent.set(Event::JoystickOneRight,value);
  }
  
  // UP
  value=
    ((cond.joyy < JOYMIN) || ((~cond.buttons) & CONT_DPAD_UP))
    ? 1 : 0;
  if(port == 0)
  {
    theEvent.set(Event::JoystickZeroUp,value);
  }
  else
  {
    theEvent.set(Event::JoystickOneUp,value);
  }
  
  // DOWN
  value=
    ((cond.joyy > JOYMAX) || ((~cond.buttons) & CONT_DPAD_DOWN))
    ? 1 : 0;
  if(port == 0)
  {
    theEvent.set(Event::JoystickZeroDown,value);
  }
  else
  {
    theEvent.set(Event::JoystickOneDown,value);
  }

  // FIRE
  value=((~cond.buttons) & CONT_A) ? 1 : 0;
  if(port == 0)
  {
    theEvent.set(Event::JoystickZeroFire,value);
  }
  else
  {
    theEvent.set(Event::JoystickOneFire,value);
  }
}

void handleBooster(cont_cond_t cond,int port)
{
  /*
   * handle the atari booster grip controller extension,
   * basis is joystick
   */
  
  Int32 value;

  if(cond.rtrig == 255) return;
  
  // FIRE
  value=((~cond.buttons) & CONT_X) ? 1 : 0;
  if(port == 0)
  {
    theEvent.set(Event::BoosterGripZeroTrigger,value);
  }
  else
  {
    theEvent.set(Event::BoosterGripOneTrigger,value);
  }
  
  // FIRE
  value=((~cond.buttons) & CONT_B) ? 1 : 0;
  if(port == 0)
  {
    theEvent.set(Event::BoosterGripZeroTrigger,value);
  }
  else
  {
    theEvent.set(Event::BoosterGripOneTrigger,value);
  }
  
}


void handlePaddle(cont_cond_t cond,int port)
{
  /*
   * handle the atari paddle controllers
   */
  
  #define POTMIN  2000
  #define POTMAX 16000

  static bool relmode[4]={false,false,false,false};
  static uInt32 potpos[4]={0 ,0 ,0, 0};

  uInt32 value;
  
  if((~cond.buttons) & CONT_DPAD_UP)
    relmode[port]=true;

  if((~cond.buttons) & CONT_DPAD_DOWN)
    relmode[port]=false;

  
  if(relmode[port])
  {
    potpos[port] = ((256-cond.joyx) << 6) + 1024; 
  }
  else
  {
    // potpos[port] = 74 * (255 - cond.joyx);
    potpos[port]=((POTMAX-POTMIN)/255)*(255-cond.joyx)+POTMIN;
  }

  
  // FIRE
  value=((~cond.buttons) & CONT_A) ? 1 : 0;

  switch(port)
  {
   case 0:
    printxy("             ",0,10,0,0);
    printint(potpos[port],0,10,0xffff,0);
    theEvent.set(Event::PaddleZeroResistance, potpos[port]);
    theEvent.set(Event::PaddleZeroFire,value);
    break;
   case 1:
    theEvent.set(Event::PaddleOneResistance, potpos[port]);
    theEvent.set(Event::PaddleOneFire,value);
    break;
   case 2:
    theEvent.set(Event::PaddleTwoResistance, potpos[port]);
    theEvent.set(Event::PaddleTwoFire,value);
    break;
   case 3:
    theEvent.set(Event::PaddleThreeResistance, potpos[port]);
    theEvent.set(Event::PaddleThreeFire,value);
    break;
   default:
    break;
  }
}

void handleKeypad(kbd_cond_t cond,int port)
{
  /*
   * handle the atari keypad controllers
   */

  uint8 keycodes[] =
  {
      KBD_KEY_1, KBD_KEY_2, KBD_KEY_3,
      KBD_KEY_Q, KBD_KEY_W, KBD_KEY_E,
      KBD_KEY_A, KBD_KEY_S, KBD_KEY_D,
      KBD_KEY_Z, KBD_KEY_X, KBD_KEY_C,
      
      KBD_KEY_8, KBD_KEY_9, KBD_KEY_0,
      KBD_KEY_I, KBD_KEY_O, KBD_KEY_P,
      KBD_KEY_K, KBD_KEY_L, KBD_KEY_SEMICOLON,
      KBD_KEY_COMMA, KBD_KEY_PERIOD, KBD_KEY_SLASH
  };
  
  
  Event::Type keyEvents[] = 
  {
      Event::KeyboardZero1, Event::KeyboardZero2, Event::KeyboardZero3,
      Event::KeyboardZero4, Event::KeyboardZero5, Event::KeyboardZero6,
      Event::KeyboardZero7, Event::KeyboardZero8, Event::KeyboardZero9,
      Event::KeyboardZeroStar, Event::KeyboardZero0, Event::KeyboardZeroPound,
      
      Event::KeyboardOne1, Event::KeyboardOne2, Event::KeyboardOne3,
      Event::KeyboardOne4, Event::KeyboardOne5, Event::KeyboardOne6,
      Event::KeyboardOne7, Event::KeyboardOne8, Event::KeyboardOne9,
      Event::KeyboardOneStar, Event::KeyboardOne0, Event::KeyboardOnePound
  };

  int i,j;
  
  for(i=port*12;i<(port+1)*12;i++)
  {
    int keypressed;
    for(j=0,keypressed=0;j<6;j++)
    {
      if(keycodes[i] == cond.keys[j])
      {
	keypressed=1;
      }
    }
    
    if(keypressed)
    {
      theEvent.set(keyEvents[i],1);
    }
    else
    {
      theEvent.set(keyEvents[i],0);
    }
  }
}



void handleConsole(cont_cond_t cond)
{
  /*
   * handle the atari console switches
   */

  static cont_cond_t lastcond;

  if((~cond.buttons) & CONT_X)
  {
    theEvent.set(Event::ConsoleSelect,1);
  }
  else
  {
    theEvent.set(Event::ConsoleSelect,0);
  }
  
  if((~cond.buttons) & CONT_Y)
  {
    theEvent.set(Event::ConsoleReset,1);
  }
  else
  {
    theEvent.set(Event::ConsoleReset,0);
  }

  if((((~cond.buttons) & CONT_B) != ((~lastcond.buttons) & CONT_B))
     && ((~cond.buttons) & CONT_B))
  {
    if(theEvent.get(Event::ConsoleBlackWhite))
    {
      theEvent.set(Event::ConsoleColor,1);
      theEvent.set(Event::ConsoleBlackWhite,0);
    }
    else
    {
      theEvent.set(Event::ConsoleColor,0);
      theEvent.set(Event::ConsoleBlackWhite,1);
    }
  }
  lastcond=cond;
}



void setupDC(int vidMode)
{
  /*
   * setting up the basic video stuff on DC
   */
  
  vid_init(vid_check_cable(),vidMode,PM_RGB555);

  if(vidMode == DM_640x480)
  {
    DC_screen_size_X=640;
    DC_screen_size_Y=480;
  }
  else
  {
    DC_screen_size_X=320;
    DC_screen_size_Y=240;
  }
  topLeft=(uint32)vram_l + theXOffset +
    (((120 - (theHeight-theYOffset)/2) & ~1) * DC_screen_size_X);
}



/*
 * subroutines used for the Stella part
 */

void setupPalette()
{ 
  /*
   * set up the color palette for Stella
   */
  
  const uInt32* palette = theConsole->mediaSource().palette();
  for(uInt32 t = 0; t < 256; t+=2)
  {
    // breaking RGB888 down to RGB555
    uInt16 RGB=
      ((palette[t] & 0x00f80000) >> 9) | // 8 if RGB565
      ((palette[t] & 0x0000f800) >> 6) | // 5 if RGB565 (also fc)
      ((palette[t] & 0x000000f8) >> 3);
    theColorTable[t]   = RGB << 16 | RGB;
    theColorTable[t+1] = RGB << 16 | RGB;
  }
}


void setDifficulty(int player,int diff)
{
  /*
   * set the difficulty
   * player: 0=left; 1=right
   * diff:   0=A;    1=B
   */
  if(!player)
  {
    theEvent.set(Event::ConsoleLeftDifficultyA,!diff);
    theEvent.set(Event::ConsoleLeftDifficultyB, diff);
  }
  else
  {
    theEvent.set(Event::ConsoleRightDifficultyA,!diff);
    theEvent.set(Event::ConsoleRightDifficultyB, diff);
  }
}

int getDifficulty(int player)
{
  /*  
   * set the difficulty
   * player: 0=left; 1=right
   * returns 0=A;    1=B
   */
  if(!player)
  {
    return theEvent.get(Event::ConsoleLeftDifficultyB);
  }
  else
  {
    return theEvent.get(Event::ConsoleRightDifficultyB);
  }
}


void updateDisplay(MediaSource& mediaSource)
{
  /*
   * update the display of the Dreamcast from Stella's data
   */
  
  uInt8* currentFrame = mediaSource.currentFrameBuffer();
  uInt8* previousFrame = mediaSource.previousFrameBuffer();
  uInt32* frameBuffer;
  
  for(uInt16 y = 0; y < theHeight; y++)
  {
    frameBuffer = (uInt32*)topLeft + (y*DC_screen_size_X) /2;
    for(uInt16 x = 0; x < theWidth; x++)
    {
      if((*currentFrame != *previousFrame) || theRedrawEntireFrameFlag)
      {
	*frameBuffer=theColorTable[*currentFrame];
      }
      currentFrame++;
      previousFrame++;
      frameBuffer++;
    }
  }
  theRedrawEntireFrameFlag=false;
}




int getControllerByName(const char *name)
{
  /*
   * convert controller type from name to internally used number
   * 
   * name: name of controller type
   */
  
  if(!strncmp(name,"Joystick",3)) return CONTROLLER_JOYSTICK;
  if(!strncmp(name,"Paddles",3))  return CONTROLLER_PADDLES;
  if(!strncmp(name,"Booster",3))  return CONTROLLER_BOOSTER;
  if(!strncmp(name,"Keypad",3))   return CONTROLLER_KEYPAD;
  if(!strncmp(name,"Driving",3))  return CONTROLLER_DRIVING;
  if(!strncmp(name,"Lightgun",3)) return CONTROLLER_LIGHTGUN;
  if(!strncmp(name,"Mindlink",3)) return CONTROLLER_MINDLINK;
  return CONTROLLER_NONE;
}


void clear_cond(cont_cond_t *cond)
{
  cond->buttons = 0;
  cond->rtrig   = 0;
  cond->ltrig   = 0;
  cond->joyx    = 128;
  cond->joyy    = 128;
  cond->joy2x   = 128;
  cond->joy2y   = 128;
}


void cont_handle_get_cond(int port,cont_cond_t *retval)
{
  static uInt8 mcont[4]={0,0,0,0};

  if(!mcont[port])
  {
    mcont[port] = maple_create_addr(port,0);
  }
  
  if(mcont[port])
  {
    if(!cont_get_cond(mcont[port],retval))
    {
      return;
    }
  }

  retval->buttons = 0xffff;
  retval->rtrig   = 0;
  retval->ltrig   = 0;
  retval->joyx    = 128;
  retval->joyy    = 128;
  retval->joy2x   = 128;
  retval->joy2y   = 128;
  return;
}


void handleEvents(int left,int right)
{
  /*
   * This routine should be called regularly to handle events
   * 
   * left,right: controller type on left and right port
   */
  
  cont_cond_t cond1,cond2,cond3,cond4;
  static cont_cond_t last1;
  static uInt8 mcontk;
  static kbd_cond_t condk;
  
  if((left == CONTROLLER_KEYPAD) || (right == CONTROLLER_KEYPAD))
  {
    if(!mcontk)
    {
      mcontk=maple_kb_addr();
    }
    kbd_get_cond(mcontk, &condk);
  }
  
  
    cont_handle_get_cond(0,&cond1);
    cont_handle_get_cond(1,&cond2);
  

  if(right == CONTROLLER_PADDLES)
  {
    cont_handle_get_cond(2,&cond3);
    cont_handle_get_cond(3,&cond4);
  }

  if((cond1.rtrig == 255) || (~cond1.buttons & CONT_C)) handleConsole(cond1);

#if 0
  if(portSwapAllowed)
  {
    if ((cond1.ltrig == 255) && (last1.ltrig != 255))
    {
      portsSwaped ^= true;
    }
    if ((~cond1.buttons & CONT_Z) && (last1.buttons & CONT_Z))
    {
      portsSwaped ^= true;
    }
  }
#else
  if((portSwapAllowed) &&
     (((cond1.ltrig == 255) && (last1.ltrig != 255)) ||
      ((~cond1.buttons & CONT_Z) && (last1.buttons & CONT_Z))))
  {
    portsSwaped ^= true;
  }
#endif
  last1=cond1;

  if(!portsSwaped)
  {
    switch(left)
    {
     case CONTROLLER_JOYSTICK:
      handleJoystick(cond1,0);
      break;
     case CONTROLLER_PADDLES:
      handlePaddle(cond1,0);
      handlePaddle(cond2,1);
      break;
     case CONTROLLER_BOOSTER:
      handleJoystick(cond1,0);
      handleBooster(cond1,0);
      break;
     case CONTROLLER_KEYPAD:
      handleKeypad(condk,0);
      break;
     case CONTROLLER_DRIVING:
      handleJoystick(cond1,0);
      break;
     case CONTROLLER_LIGHTGUN:
      break;
     case CONTROLLER_MINDLINK:
     default:
      break;
    }
    
    switch(right)
    {
     case CONTROLLER_JOYSTICK:
      handleJoystick(cond2,1);
      break;
     case CONTROLLER_PADDLES:
      handlePaddle(cond3,2);
      handlePaddle(cond4,3);
      break;
     case CONTROLLER_BOOSTER:
      handleJoystick(cond2,1);
      handleBooster(cond2,1);
      break;
     case CONTROLLER_KEYPAD:
      handleKeypad(condk,1);
      break;
     case CONTROLLER_DRIVING:
      handleJoystick(cond2,1);
      break;
     case CONTROLLER_LIGHTGUN:
      break;
     case CONTROLLER_MINDLINK:
      break;
     default:
      break;
    }
  }
  else
  {
    switch(left)
    {
     case CONTROLLER_JOYSTICK:
      handleJoystick(cond2,0);
      break;
     case CONTROLLER_PADDLES:
      handlePaddle(cond3,0);
      handlePaddle(cond4,1);
      break;
     case CONTROLLER_BOOSTER:
      handleJoystick(cond2,0);
      handleBooster(cond2,0);
      break;
     case CONTROLLER_KEYPAD:
      handleKeypad(condk,0);
      break;
     case CONTROLLER_DRIVING:
      handleJoystick(cond2,0);
      break;
     case CONTROLLER_LIGHTGUN:
      break;
     case CONTROLLER_MINDLINK:
     default:
      break;
    }
    
    switch(right)
    {
     case CONTROLLER_JOYSTICK:
      handleJoystick(cond1,1);
      break;
     case CONTROLLER_PADDLES:
      handlePaddle(cond1,2);
      handlePaddle(cond2,3);
      break;
     case CONTROLLER_BOOSTER:
      handleJoystick(cond1,1);
      handleBooster(cond1,1);
      break;
     case CONTROLLER_KEYPAD:
      handleKeypad(condk,1);
      break;
     case CONTROLLER_DRIVING:
      handleJoystick(cond1,1);
      break;
     case CONTROLLER_LIGHTGUN:
      break;
     case CONTROLLER_MINDLINK:
      break;
     default:
      break;
    }
  }
}




uint16 hex16(const char *s)
{
  return (uint16) strtol (s, NULL, 16);
}

uint32 requestFile(uint8 *image,ssize_t shouldsize)
{
  int done=0,pospointer,maxroms;
  unsigned int *cmd;
  
  uint32 buttons;

  uint16 textbgcol,textfgcol,barfgcol,barbgcol;

  if(strlen(SELECT_MENUTEXT_FGCOL))
  {
    textfgcol=(((hex16(SELECT_MENUTEXT_FGCOL) & 0x00f) << 1) |
	       ((hex16(SELECT_MENUTEXT_FGCOL) & 0x0f0) << 2) |
	       ((hex16(SELECT_MENUTEXT_FGCOL) & 0xf00) << 3));
  }
  else
  {
    textfgcol=0x7fff;
  }
  
  if(strlen(SELECT_MENUTEXT_BGCOL))
  {
    textbgcol=(((hex16(SELECT_MENUTEXT_BGCOL) & 0x00f) << 1) |
	       ((hex16(SELECT_MENUTEXT_BGCOL) & 0x0f0) << 2) |
	       ((hex16(SELECT_MENUTEXT_BGCOL) & 0xf00) << 3));
  }
  else
  {
    textbgcol=0x0000;
  }
  
  if(strlen(SELECT_MENUBAR_FGCOL))
  {
    barfgcol=(((hex16(SELECT_MENUBAR_FGCOL) & 0x00f) << 1) |
	      ((hex16(SELECT_MENUBAR_FGCOL) & 0x0f0) << 2) |
	      ((hex16(SELECT_MENUBAR_FGCOL) & 0xf00) << 3));
  }
  else
  {
    barfgcol=0x7fff;
  }

  if(strlen(SELECT_MENUBAR_BGCOL))
  {
    barbgcol=(((hex16(SELECT_MENUBAR_BGCOL) & 0x00f) << 1) |
	      ((hex16(SELECT_MENUBAR_BGCOL) & 0x0f0) << 2) |
	      ((hex16(SELECT_MENUBAR_BGCOL) & 0xf00) << 3));
  }
  else
  {
    barbgcol=0x000f;
  }

  select_one_from_list(romname, 0, CONT_A | CONT_Y,
		       &actualgame,&topoffsetgame,
		       SELECT_MENU_XPOS,SELECT_MENU_YPOS,
		       SELECT_MENU_WIDTH,SELECT_MENU_HEIGHT,
		       textfgcol,textbgcol,barfgcol,barbgcol);

  timer_sleep(20);
  buttons=get_cont_buttons(0);
  
  if(buttons & CONT_X)
  {
    // activate debug mode
    emuVidMode=DM_640x480;
  }
  else
  {
    emuVidMode=DM_320x240;
  }

#if (DEBUG == TRUE)
  if(buttons & CONT_Y)
  {
    // prepare exiting
    return(0);
  }
#endif

#if 0  
**  fd=iso_open(get_filename(actualgame),O_RDONLY);
  if(!fd)
  {
    error_flash(4);
    return 0;
  }
  size=iso_read(fd,image,shouldsize);
  iso_close(fd);
#endif
  
//  return (load_rom(get_filename(actualgame),image,shouldsize));
  return (load_file(get_filename(actualgame),(void**)&image));
}


void difficulty_on_vmu()
{
  if(theEvent.get(Event::ConsoleLeftDifficultyA))  vmu_setup(0,1);
  if(theEvent.get(Event::ConsoleLeftDifficultyB))  vmu_setup(0,2);
  if(theEvent.get(Event::ConsoleRightDifficultyA)) vmu_setup(1,1);
  if(theEvent.get(Event::ConsoleRightDifficultyB)) vmu_setup(1,2);
}

#define ROMINFO_COLOR_BORDER     0x00e0
#define ROMINFO_COLOR_HEADER     0x03ff
#define ROMINFO_COLOR_DATA       0x001f
#define ROMINFO_COLOR_BACKGROUND 0x0000

void ri_line(char *text,int y,int fg)
{
#define ROMINFO_WIDTH 32
  char printbuf[ROMINFO_WIDTH+1];
  printxy("*                                *",10003,y,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);
  strncpy(printbuf,text,ROMINFO_WIDTH);
  printbuf[ROMINFO_WIDTH]='\0';
  if(strcmp(printbuf,""))
  {
    printxy(printbuf,10004,y,fg,ROMINFO_COLOR_BACKGROUND);
  }
  else
  {
    printxy("(not set)",10004,y,fg & ~0x4210,ROMINFO_COLOR_BACKGROUND);
  }
}


#ifdef string
void rom_info(string rommd5)
#else
void rom_info(std::string rommd5)
#endif
{
  int y=10003;
  char
  *cart_md5 =(char*)theConsole->properties().get("Cartridge.MD5").c_str(),
  *cart_manu=(char*)theConsole->properties().get("Cartridge.Manufacturer").c_str(),
  *cart_no  =(char*)theConsole->properties().get("Cartridge.ModelNo").c_str(),
  *cart_name=(char*)theConsole->properties().get("Cartridge.Name").c_str(),
  *cart_note=(char*)theConsole->properties().get("Cartridge.Note").c_str(),
  *cart_rare=(char*)theConsole->properties().get("Cartridge.Rarity").c_str(),
  *cart_type=(char*)theConsole->properties().get("Cartridge.Type").c_str(),
//  *con_diffl=(char*)theConsole->properties().get("Console.LeftDifficulty").c_str(),
//  *con_diffr=(char*)theConsole->properties().get("Console.RightDifficulty").c_str(),
  *con_contl=(char*)theConsole->properties().get("Controller.Left").c_str(),
  *con_contr=(char*)theConsole->properties().get("Controller.Right").c_str(),
  *disp_form=(char*)theConsole->properties().get("Display.Format").c_str(),
//  *disp_rate=(char*)theConsole->properties().get("Display.FrameRate").c_str(),
//  *disp_xsiz=(char*)theConsole->properties().get("Display.Height").c_str(),
//  *disp_ysiz=(char*)theConsole->properties().get("Display.Width").c_str(),
//  *disp_x   =(char*)theConsole->properties().get("Display.XStart").c_str(),
//  *disp_y   =(char*)theConsole->properties().get("Display.YStart").c_str(),
//  *emu_cpu  =(char*)theConsole->properties().get("Emulation.CPU").c_str(),
//  *emu_blan =(char*)theConsole->properties().get("Emulation.HmoveBlanks").c_str(),
//  *tia_delay=(char*)theConsole->properties().get("TIA.PlayerDelay").c_str(),
//  *timer_adj=(char*)theConsole->properties().get("Timer.Adjustment").c_str(),
  buffer[80];
  
  theRedrawEntireFrameFlag = true;
  updateDisplay(theConsole->mediaSource());
  printxy("**********************************",10003,y++,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);
  ri_line("         " DCSTELLA " " VERSION,y++,0x7000);
  ri_line("  http://our-arca.de/DCstella/",y++,0x60c0);
  ri_line("",y++,0x0000);
  ri_line("Name/Display/Rarity/PropertiesID",y++,ROMINFO_COLOR_HEADER);
  ri_line(cart_name,y++,ROMINFO_COLOR_DATA);
  ri_line(disp_form,y++,ROMINFO_COLOR_DATA);
  ri_line(cart_rare,y++,ROMINFO_COLOR_DATA);
  sprintf(buffer,"%s.pro",rommd5.substr(0,8).c_str());
  ri_line(buffer,y++,ROMINFO_COLOR_DATA);
  ri_line("",y++,0x0000);
  ri_line("Manufacturer/Model No.",y++,ROMINFO_COLOR_HEADER);
  ri_line(cart_manu,y++,ROMINFO_COLOR_DATA);
  ri_line(cart_no,y++,ROMINFO_COLOR_DATA);
  ri_line("",y++,0x0000);
  ri_line("Controller",y++,ROMINFO_COLOR_HEADER);
  sprintf(buffer,"left: %s",con_contl);
  ri_line(buffer,y++,ROMINFO_COLOR_DATA);
  sprintf(buffer,"right:%s",con_contr);
  ri_line(buffer,y++,ROMINFO_COLOR_DATA);
  ri_line("",y++,0x0000);
  ri_line("Type/Note:",y++,ROMINFO_COLOR_HEADER);
  ri_line(cart_type,y++,ROMINFO_COLOR_DATA);
  ri_line(cart_note,y++,ROMINFO_COLOR_DATA);
  printxy("*********  PRESS  START  *********",10003,y++,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);

  while(!(get_cont_buttons(0) & CONT_START))
  {
    gdrom_check_lid();
    vid_waitvbl();
  }

  while( (get_cont_buttons(0) & CONT_START))
  {
    gdrom_check_lid();
    vid_waitvbl();
  }

  vid_clear(0,0,0);
  theRedrawEntireFrameFlag = true;
  updateDisplay(theConsole->mediaSource());
}

#define CHEAT_COLOR_ACTUAL_ADDRESS     (( 0 << 10) | (20 << 5) |  0)
#define CHEAT_COLOR_ACTUAL_ODDBYTE     (( 0 << 10) | (20 << 5) |  0)
#define CHEAT_COLOR_ACTUAL_EVENBYTE    (( 0 << 10) | (15 << 5) |  0)
#define CHEAT_COLOR_MARKED_ODDBYTE     ((20 << 10) | ( 0 << 5) | 20)
#define CHEAT_COLOR_MARKED_EVENBYTE    ((15 << 10) | ( 0 << 5) | 15)
#define CHEAT_COLOR_XMARKED_ODDBYTE    ((20 << 10) | ( 0 << 5) |  0)
#define CHEAT_COLOR_XMARKED_EVENBYTE   ((15 << 10) | ( 0 << 5) |  0)
#define CHEAT_COLOR_SAVED_ADDRESS      (( 0 << 10) | ( 0 << 5) | 15)
#define CHEAT_COLOR_SAVED_ODDBYTE      (( 0 << 10) | ( 0 << 5) | 15)
#define CHEAT_COLOR_SAVED_EVENBYTE     (( 0 << 10) | ( 0 << 5) | 10)
#define CHEAT_COLOR_CURSOR             ((20 << 10) | (20 << 5) |  0)
#define CHEAT_COLOR_LOCKED             ((25 << 10) | ( 0 << 5) |  0)
#define CHEAT_COLOR_BACKGROUND         0x0000
#define CHEAT_BYTES_PER_LINE           0x10


      
void cheat_init()
{
  int i;
  for(i=0;i<0x80;i++)
  {
    cheatSavedMem[i]=0;
  }
  cheatAddress   = 0;
  cheatValue     = 0;
  cheatCursorPos = 0;
}


uint8 cheat_d2x(char *value)
{
  uint8 digit,bin=0;
  for(digit=0;digit<4;digit++)
  {
    bin <<= 2;
    switch(toupper(value[digit]))
    {
     case 'A':
      break;
     case 'B':
      bin |= 1;
      break;
     case 'X':
      bin |= 2;
      break;
     case 'Y':
      bin |= 3;
      break;
    }
  }
  return bin;
}


void cheat_x2d(int value,char *buffer)
{
  char dccodes[]="ABXY";
  int digit,t;
  buffer[4]='\0';
  for(digit=3;digit>=0;digit--)
  {
    buffer[digit] = dccodes[value % 4];
    value >>= 2;
  }
}


void cheat_get_code()
{
  char code_buffer[]="........";
  int end=0,waitforinput=0,pos=0;
  char buffer[20];
  uint32 buttons;
  
  while(!end)
  {
    gdrom_check_lid();
    vid_waitvbl();
    
    printxy(code_buffer,10021,10017,0xffff,0x000f);

    buttons=get_cont_buttons(0);
    if(!buttons)
    {
      waitforinput=1;
      if(pos > 7)
      {
	end = 1;
      }
      if(pos < 0)
      {
	end = 2;
      }
    }
    else
    {
      if(waitforinput)
      {
	waitforinput=0;
	if(buttons & CONT_A)
	  code_buffer[pos++]='A';
	if(buttons & CONT_B)
	  code_buffer[pos++]='B';
	if(buttons & CONT_X)
	  code_buffer[pos++]='X';
	if(buttons & CONT_Y)
	  code_buffer[pos++]='Y';
	 
	if(buttons & CONT_DPAD_LEFT)
	{
	  if(pos > 0)
	    code_buffer[--pos]='.';
	  else
	    pos--;
	}
      }
    }
  }
  if(end == 1)
  {
    cheatAddress   = cheat_d2x(&code_buffer[0]);
    cheatValue     = cheat_d2x(&code_buffer[4]);
    cheatCursorPos = cheatAddress & 0x7f;
  }
}


void cheat_display()
{
  int line,byte,x,y,menu_timer=0;
  uInt16 address,color;
  char buffer[80];
  uint32 buttons;
  
#define YBASE1 10006
#define YBASE2 10007

#define COLOR_A_BUTTON		((19 << 10) | ( 0 << 5) |  0)
#define COLOR_B_BUTTON		(( 0 << 10) | ( 0 << 5) | 19)
#define COLOR_X_BUTTON		((19 << 10) | (19 << 5) |  0)
#define COLOR_Y_BUTTON		(( 0 << 10) | (19 << 5) |  0)
#define COLOR_START_BUTTON	((19 << 10) | (19 << 5) | 19)
  
  vid_clear(0,0,0);
  
  /*
   * Cheat Menu:
   * DPAD - Move
   * Y - save actual to compare buffer
   * B - copy byte from compare buffer
   * A - (un)lock byte
   * 
   */
  
  y=10002;
  printxy("**********************************",10003,y++,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);
  ri_line("         " DCSTELLA " " VERSION,y++,0x7000);
  ri_line("       Cheat code monitor",y++,0x7000);
  printxy("************************************",10002,y++,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);

  y+=16;
  printxy("************************************",10002,y++,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);
  printxy("*          *           *           *",10002,y,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);
  printxy("Saved  RAM",10003,y,CHEAT_COLOR_SAVED_ADDRESS,CHEAT_COLOR_BACKGROUND);
  printxy("B:Load Byte",10026,y,COLOR_B_BUTTON,CHEAT_COLOR_BACKGROUND);
  printxy("Y:Save RAM ",10014,y++,COLOR_Y_BUTTON,CHEAT_COLOR_BACKGROUND);
  printxy("*          *           *           *",10002,y,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);
  printxy("Actual RAM",10003,y,CHEAT_COLOR_ACTUAL_ADDRESS,CHEAT_COLOR_BACKGROUND);
  printxy("A:Lock Byte",10026,y,COLOR_A_BUTTON,CHEAT_COLOR_BACKGROUND);
  printxy("Start:Exit ",10014,y++,COLOR_START_BUTTON,CHEAT_COLOR_BACKGROUND);
  printxy("************************************",10002,y++,ROMINFO_COLOR_BORDER,ROMINFO_COLOR_BACKGROUND);
  
  while(buttons != CONT_START)
  {
    gdrom_check_lid();
    vid_waitvbl();

#if 0
    cheat_x2d(cheatSavedMem[cheatCursorPos],buffer);
    printxy(buffer,10020,10024,CHEAT_COLOR_CURSOR,CHEAT_COLOR_BACKGROUND);
    cheat_x2d(cheatCursorPos+0x80,buffer);
    printxy(buffer,10016,10024,CHEAT_COLOR_CURSOR,CHEAT_COLOR_BACKGROUND);
#endif
    
    for(y=0;y<8;y++)
    {
      sprintf(buffer,"%02x: ",0x80+y*CHEAT_BYTES_PER_LINE);
      printxy(buffer,10002,YBASE1+y*2,CHEAT_COLOR_SAVED_ADDRESS,CHEAT_COLOR_BACKGROUND);
      printxy(buffer,10002,YBASE2+y*2,CHEAT_COLOR_ACTUAL_ADDRESS,CHEAT_COLOR_BACKGROUND);
      for(x=0;x<0x10;x++)
      {
	address=0x80+y*0x10+x;

	/* print out the saved memory */
	sprintf(buffer,"%02x",cheatSavedMem[address-0x80]);
	if(address == cheatAddress)
	{
	  color = CHEAT_COLOR_LOCKED;
	}
	else
	{
	  color =
	    (x%2) ? CHEAT_COLOR_SAVED_ODDBYTE : CHEAT_COLOR_SAVED_EVENBYTE;
	}
	printxy(buffer,10006+x*2,YBASE1+y*2,
		color, CHEAT_COLOR_BACKGROUND);
	
	/* print out the actual memory */
	sprintf(buffer,"%02x",theConsole->system().peek(address));
	if(address == 0x80+cheatCursorPos)
	{
	  color = CHEAT_COLOR_CURSOR;
	}
	else
	{
	  if(theConsole->system().peek(address) == cheatSavedMem[address-0x80] - 1)
	  {
	    color =
	      (x%2) ? CHEAT_COLOR_XMARKED_ODDBYTE : CHEAT_COLOR_XMARKED_EVENBYTE;
	  }
	  else if(theConsole->system().peek(address) != cheatSavedMem[address-0x80])
	  {
	    color =
	      (x%2) ? CHEAT_COLOR_MARKED_ODDBYTE : CHEAT_COLOR_MARKED_EVENBYTE;
	  }
	  else
	  {
	    color =
	      (x%2) ? CHEAT_COLOR_ACTUAL_ODDBYTE : CHEAT_COLOR_ACTUAL_EVENBYTE;
	  }
	}
	
	printxy(buffer,10006+x*2,YBASE2+y*2,
		color, CHEAT_COLOR_BACKGROUND);
      }
    }
    
    buttons=get_cont_buttons(0);

    if(menu_timer <= 0)
    {
      menu_timer = 15;

      if(buttons & CONT_DPAD_DOWN)
	cheatCursorPos+=0x10;
      if(buttons & CONT_DPAD_UP)
	cheatCursorPos-=0x10;
      if(buttons & CONT_DPAD_RIGHT)
	cheatCursorPos++;
      if(buttons & CONT_DPAD_LEFT)
	cheatCursorPos--;

      if(cheatCursorPos < 0x00)
	cheatCursorPos += 0x80;
      if(cheatCursorPos > 0x7f)
	cheatCursorPos -= 0x80;
      
      if(buttons & CONT_Y)
      {
	for( byte = 0; byte < 0x80; byte++ )
	{
	  cheatSavedMem[byte] = theConsole->system().peek(byte+0x80);
	}
      }
      
      if(buttons & CONT_B)
      {
	theConsole->system().poke(cheatCursorPos+0x80,cheatSavedMem[cheatCursorPos]);
      }
            
      if(buttons & CONT_A)
      {
	if( cheatAddress != 0x80 + cheatCursorPos )
	{
	  cheatAddress = 0x80 + cheatCursorPos;
	  cheatValue   = cheatSavedMem[cheatCursorPos];
	}
	else
	{
	  cheatAddress = 0;
	}
      }
    }
    
    if(!buttons)
      menu_timer=0;
    else
      menu_timer--;
  }

  while(get_cont_buttons(0))
  {
    gdrom_check_lid();
    vid_waitvbl();
  }
  
  vid_clear(0,0,0);
  theRedrawEntireFrameFlag = true;
  updateDisplay(theConsole->mediaSource());
}




int main(int argc, char* argv[])
{
  uInt8* romimage;
  uInt32 romsize;
  int  controller_left,controller_right;
  bool start_pressed=false;
  bool menu_active=false;
  bool autorun_mode=false;
  int  menu_line=0;
  int  menu_timer=0;
  void *targa_image=NULL;
  void *modplayer=NULL;
  void *mod=NULL;
  int  modplayer_size;
  int  mod_start;
  int  mod_size;
  void *tmpvoid;
  
#ifdef string
  string rommd5;
#else
  std::string rommd5;
#endif

  dc_setup_quiet(DM_640x480,PM_RGB555);
  serial_init(115200); // dc_setup sets to 57600
  printf(DCSTELLA " " VERSION " (Build %d)...\n",BUILDNUMBER());
  
  setupDC(DM_640x480);
  vid_empty();

  load_config(CONFIG_FILE);

#if (DEBUG == TRUE)
// override some settings from dcstella.ini
//  config_set_key_value("select.audio.file","/mods/aws_adwh.dxm");
  config_set_key_value("select.audio.player","");
//  config_set_key_value("autorun.file","/roms/a7b96a81.bin");
  config_set_key_value("select.menutext.fgcolor","0xff7");
  config_set_key_value("select.menutext.bgcolor","0x000");
  config_set_key_value("select.menubar.fgcolor","0x0ff");
  config_set_key_value("select.menubar.bgcolor","0x008");
  config_set_key_value("cheat.allowed","yes");
#endif
  
  load_file(TITLE_IMAGE_FILE,&targa_image);
  display_targa(targa_image,
		TITLE_IMAGE_XPOS,
		TITLE_IMAGE_YPOS,
		PM_RGB555);
  free(targa_image);
  targa_image=NULL;

  if(strlen(AUTORUN_FILE))
  {
    autorun_mode = true;
  }

  if(!strcasecmp(CHEAT_ALLOWED,"yes"))
  {
    cheatAllowed = true;
  }
  
  /* load the 8x8 font to memory */
  load_file(CHARSET_8X8_FILE,(void**)&charset8x8);
    
  /* shall the 8x16 font be loaded? */
  if(strlen(CHARSET_8X16_FILE))
  {
    /* load the 8x16 font to memory */
    if(!load_file(CHARSET_8X16_FILE,(void**)&charset8x16))
    {
      charset8x8to8x16();
    }
  }
  else
  {
    charset8x8to8x16();
  }
  
  /* load the background music */
  if(modplayer_size=load_file(SELECT_AUDIO_PLAYER,&modplayer))
  {
    /* no player no data needed */
    if(mod_size=load_file(SELECT_AUDIO_FILE,&mod))
    {
      /* check for sane start address */
      if((mod_start=SELECT_AUDIO_START) < modplayer_size)
      {
	/* fallback: start address of s3mplay */
	mod_start=0x10000;
      }
      /* module data should not be overwritten :) */
//      snd_load(mod,mod_start,mod_size);
    }
  }
  
  if(!autorun_mode)
  {
    /* load the menu screen to memory */
    load_file(SELECT_IMAGE_FILE,&targa_image);
  
    /* check for disc swap */
    if(get_cont_buttons(0) & CONT_START)
    {
      printxy("                                                 ",10015,10008,0xffff,0);
      printxy(" Insert CD containing ROM images and press start ",10015,10009,0xffff,0);
      printxy("                                                 ",10015,10010,0xffff,0);
      /* wait for start release, pressed again and cdrom read by drive */
      while  (get_cont_buttons(0) & CONT_START)  vid_waitvbl();
      while(!(get_cont_buttons(0) & CONT_START)) vid_waitvbl();
      while( (gdrom_check_drive() & 0x0f) != 1 ) vid_waitvbl();
      /* re-init, just in case */
      cdrom_init();
      iso_init();
    }
    generate_roms_list(ROMS_DIR,ROMS_LIST_FILE);
  }
  actualgame = 0;
  topoffsetgame = 0;
  
  if(!strcmp(PROPERTIES_DIR,""))
  {
    propertiesSet.load(PROFILES_DIR,&Console::defaultProperties(),true);
  }
  else
  {
    propertiesSet.load(PROPERTIES_DIR,&Console::defaultProperties(),true);
  }
    
  timer_init();

  
  while(true)
  {
    setupDC(DM_640x480);
    vmu_setup(0,0);

    snd_stop_arm();
    if((mod_size) && (modplayer_size))
    {
#if (DEBUG == TRUE)
      printf("playersize=%d\nmodstart  =0x%x\nmodsize   =%d\n",
	     modplayer_size,mod_start,mod_size);
#endif
      snd_load(mod,mod_start,mod_size);
      snd_load_arm(modplayer,modplayer_size);
    }
    
    romimage = new uInt8[MAXROMSIZE];
    if(autorun_mode)
    {
//      romsize = load_rom(AUTORUN_FILE,romimage,MAXROMSIZE);
      romsize = load_file(AUTORUN_FILE,(void**)&romimage);
      emuVidMode=DM_320x240;
    }
    else
    {
      display_targa(targa_image,
		    SELECT_IMAGE_XPOS,
		    SELECT_IMAGE_YPOS,
		    PM_RGB555);
      romsize = requestFile(romimage,MAXROMSIZE);
      if(!romsize) break;
    }
    
    rommd5=MD5(romimage,romsize);
//    loadProperties(rommd5);

    // Create a sound object for use with the console
    SoundDC sound;
    
    // Create the 2600 game console TODO1.2
    theConsole = new Console(romimage, romsize, DCSTELLA " " VERSION,
			     theEvent, propertiesSet, sound);
    
    theHeight    = theConsole->mediaSource().height();
    theWidth     = theConsole->mediaSource().width();
    theScanlines = theConsole->mediaSource().scanlines();
    theXOffset   = atoi((char*)theConsole->
			properties().get("Display.XStart").c_str());
    theYOffset   = atoi((char*)theConsole->
			properties().get("Display.YStart").c_str());

    setupDC(emuVidMode);
    vid_clear(0,0,0);
    cheat_init();

#if 0
    // not necessary any more...
    // setup is so quick now, message isn't seen anyway
    printxy("====================",10010,10008,0x7000,0x0000);
    printxy("=  " DCSTELLA " " VERSION "  =",10010,10009,0x7000,0x0000);
    printxy("====================",10010,10010,0x7000,0x0000);
    printxy("Setting up system...",10010,10013,0x0007,0x0000);
#endif
    maple_rescan_bus(1);
    
    setupPalette();

    // set Console switches to some sane settings, just in case
    theEvent.set(Event::ConsoleColor,1);
    theEvent.set(Event::ConsoleBlackWhite,0);
    theEvent.set(Event::ConsoleReset,0);
    theEvent.set(Event::ConsoleSelect,0);
    setDifficulty(0,0);
    setDifficulty(1,0);
    difficulty_on_vmu();
    
    // Free the image since we don't need it any longer
    delete[] romimage;

    controller_left  = getControllerByName((char*)theConsole->
				properties().get("Controller.Left").c_str());
    controller_right = getControllerByName((char*)theConsole->
				properties().get("Controller.Right").c_str());


    if(!strcmp("PAL",
	       (char*)theConsole->properties().get("Display.Format").c_str()))
    {
      // Dirty hack for better display of PAL games
      theYOffset-=50;
      theScanlines+=30;
    }
    
    theRedrawEntireFrameFlag = true;
    theQuitIndicator = false;
    menu_active = false;
    menu_line = 1;
    portSwapAllowed = false;
    portsSwaped = 0;

#if (DEBUG == TRUE)
#if 0
    // for measuring the placement on the screen
    for(int i=0;i<DC_screen_size_X;i++)
    {
      vram_s[i]=0x7000;
      vram_s[DC_screen_size_X*(DC_screen_size_Y-2)+i]=0x7000;
    }
    
    for(int i=0;i<DC_screen_size_Y;i++)
    {
      vram_s[i*DC_screen_size_X]=0x7000;
      vram_s[i*(DC_screen_size_X+1)]=0x7000;
      vram_s[((i+1)*DC_screen_size_X)-1]=0x7000;
    }
#endif
#endif
    
    while(!theQuitIndicator)
    {
      if((!(get_cont_buttons(0) & CONT_START)) && start_pressed)
      {
	menu_active = !menu_active;
	if(menu_active)
	  sound.mute(true);
	else
	{
	  sound.mute(false);
	  maple_rescan_bus(1);
	}
      }
      start_pressed=(get_cont_buttons(0) & CONT_START);

	
      if(!menu_active)
      {

	timer_prime(TMU0,61);
	timer_start(TMU0);
	
	// Draw the frame and handle events
	if (emuVidMode == DM_640x480) 
	{
	  vid_border_color(128,0,0); // red
	  theConsole->mediaSource().update();
	  vid_border_color(128,128,0); // cyan
	  updateDisplay(theConsole->mediaSource());
	  vid_border_color(0,0,128); // blue
	  handleEvents(controller_left,controller_right);
	  vid_border_color(128,0,128); // blue
	  gdrom_check_lid();
	  vid_border_color(0,0,0); //black
	}
	else
	{
	  theConsole->mediaSource().update();
	  updateDisplay(theConsole->mediaSource());
	  handleEvents(controller_left,controller_right);
	  gdrom_check_lid();
	}
//	printhex(*(uint32*)0xa080dff0,0,0,0xffff,0x0000);
	
	if(cheatAddress > 0)
	  theConsole->system().poke(cheatAddress,cheatValue);
	timer_stop(TMU0);

	if(!timer_clear(TMU0))
	{
	  // Still time to waste? Wait for vbl...
	  vid_waitvbl();
	}
      }
      else /* menu */
      {
	char pmb[21],cheatcode[9];
	uint32 buttons;

	gdrom_check_lid();
	vid_waitvbl();
	//       12345678901234567890
	printxy("====================",10010,10008,0x7000,0x0000);
	printxy("=  " DCSTELLA " " VERSION "  =",10010,10009,0x7000,0x0000);
	printxy("====================",10010,10010,0x7000,0x0000);
	//           12345678901234567890
	if(!autorun_mode)
	{
	  printxy(" Return to ROM Menu ",10010,10012,
		  0xffff,(menu_line == 0) ? 0x000f : 0);
	}
	sprintf(pmb," Difficulty Left: %c ",'A'+getDifficulty(0));
	printxy(pmb,10010,10013,
		0xffff,(menu_line == 1) ? 0x000f : 0);
	sprintf(pmb," Difficulty Right:%c ",'A'+getDifficulty(1));
	printxy(pmb,10010,10014,
		0xffff,(menu_line == 2) ? 0x000f : 0);

	sprintf(pmb," Port swap %8s ",
		portSwapAllowed ? "enabled ":"disabled");
	printxy(pmb,10010,10015,
		0xffff,(menu_line == 3) ? 0x000f : 0);

	printxy(" Turn Power Off/On  ",10010,10016,
		0xffff,(menu_line == 4) ? 0x000f : 0);

	if(cheatAllowed)
	{
	  if(cheatAddress)
	  {
	    cheat_x2d(cheatAddress,&cheatcode[0]);
	    cheat_x2d(cheatValue,  &cheatcode[4]);
	  }
	  else
	  {
	    strcpy(cheatcode,"disabled");
	  }
	  sprintf(pmb," Cheatcode:%8s ", cheatcode);
	  printxy(pmb,10010,10017,
		  0xffff,(menu_line == 5) ? 0x000f : 0);
	}
	
	printxy(" ROM properties     ",10010,10018,
		0xffff,(menu_line == 6) ? 0x000f : 0);
	
	buttons=get_cont_buttons(0);

	if(menu_timer <= 0)
	{
	  menu_timer = 15;

	  if(buttons & CONT_DPAD_DOWN)
	  {
	    menu_line++;
	    if((menu_line == 5) && !cheatAllowed)
	      menu_line++;
	  }
	  if(buttons & CONT_DPAD_UP)
	  {	    
	    menu_line--;
	    if((menu_line == 5) && !cheatAllowed)
	      menu_line--;
	  }
	  

#if 0
	  if(buttons & CONT_X)
	  {
	    unsigned long i;
	    for(i=0xa080e000;i<0xa080e800;i+=4)
	    {
	      printf("%08x",*(unsigned long*)i);
	    }
	    printf("\n");
	  }
#endif
	  
	  if(menu_line > 6) menu_line = 6;
	  if(autorun_mode)
	  {
	    if(menu_line < 1) menu_line = 1;
	  }
	  else
	  {
	    if(menu_line < 0) menu_line = 0;
	  }
	  
	  
	  if(!(buttons & (CONT_DPAD_UP | CONT_DPAD_DOWN)))
	  {
	    switch(menu_line)
	    {
	     case 0: // Return to ROM menu
	      if(buttons & (CONT_A | CONT_B | CONT_START))
	      {
		theQuitIndicator = true;
	      }
	      break;
	     case 1: // Difficulty left
	     case 2: // Difficulty right
	      if((buttons & (CONT_A | CONT_B)) != (CONT_A | CONT_B))
	      {
		if(buttons & CONT_A)
		{
		  setDifficulty(menu_line-1,0);
		}
		if(buttons & CONT_B)
		{
		  setDifficulty(menu_line-1,1);
		}
	      }
	      difficulty_on_vmu();
	      break;
	     case 3: // Right Joystick
	      if((buttons & (CONT_A | CONT_B)) != (CONT_A | CONT_B))
	      {
		if(buttons & CONT_A)
		{
		  portSwapAllowed=true;
		}
		if(buttons & CONT_B)
		{
		  portSwapAllowed=false;
		  portsSwaped=false;
		}
	      }
	      break;
	     case 4: // Reset ROM
	      if(buttons & (CONT_START | CONT_A | CONT_B))
	      {
		theConsole->system().reset();
		while(get_cont_buttons(0) & (CONT_START | CONT_A | CONT_B))
		  vid_waitvbl();
		sound.mute(false);
		menu_active=false;
	      }
	      break;
	     case 5: // Cheat System
	      if(buttons & CONT_A)
	      {
		cheat_get_code();
	      }
	      else if(buttons & CONT_B)
	      {
		cheatAddress = 0;
	      }
	      else if(buttons & CONT_X)
	      {
		cheat_display();
		sound.mute(false);
		menu_active=false;
	      }
	      break;
	     case 6: // ROM info
	      if(buttons & (CONT_A | CONT_B))
	      {
		rom_info(rommd5);
	      }
	      break;
	    }
	  }
	}
	
	if(!buttons)
	  menu_timer=0;
	else
	  menu_timer--;

#if (DEBUG == TRUE)
	if(buttons & CONT_Y)
	{
	  vid_border_color(0,0,0);
	  vid_empty();
	  dc_setup_quiet(DM_640x480,PM_RGB565);
	  serial_init(115200);
	  return 0;
	}
#endif
	theRedrawEntireFrameFlag = true;
      }
    }
    vid_border_color(0,0,0);
    vid_empty();
//    vid_clear(0,0,0);
    
    delete theConsole;
  }
  snd_init();
  dc_setup_quiet(DM_640x480,PM_RGB565); 
  serial_init(115200); // dc_setup sets to 57600
  /* clean up a bit */
    
  return(0);
}
