/*
 * This file is part of the DCStella project
 * location: dcstella/core/screen.c
 * written by SvOlli 
 */

#include <dream.h>
#include "screen.h"
#include "discio.h"
#include "o/compilemode.h"

//uint8 charset8x8[CHARSET8x8_SIZE];
//uint8 charset8x16[CHARSET8x16_SIZE];
uint8 *charset8x8;
uint8 *charset8x16;

void charset8x8to8x16()
{
  /*
   * create the 8x16 charset by converting the 8x8 one
   */
  
  int i;
  charset8x16=(char*)malloc(1536);
  for(i=0;i<CHARSET8x8_SIZE;i++)
  {
    charset8x16[i*2]=charset8x8[i];
    charset8x16[i*2+1]=charset8x8[i];
  }
}

void draw_char_s16(int charn,int x,int y,uint16 fg,uint16 bg)
{
  /*
   * draw character with single height (8x8) on 16bit framebuffer
   * 
   * charn: char to draw
   * x:     xpos in pixels, if >= 10000 in characters
   * y:     ypos in pixels, if >= 10000 in characters
   * fg:    foreground color
   * bg:    background color
   */
  int line,pixel;

  if((charn < ' ') || (charn > '~')) charn=' ';

  for(line = 0; line < 8; line++)
  {
    for(pixel = 0; pixel < 8; pixel++)
    {
      if((1 << (7-pixel) & charset8x8[(charn-32)*8+line]))
      {
        vram_s[(y+line)*DC_screen_size_X+(x+pixel)]=fg;
      }
      else
      {
        vram_s[(y+line)*DC_screen_size_X+(x+pixel)]=bg;
      }
    }
  }
}

void draw_char_d16(int charn,int x,int y,uint16 fg,uint16 bg)
{
  /*
   * draw character with double height (8x16) on 16bit framebuffer
   * 
   * charn: char to draw
   * x:     xpos in pixels, if >= 10000 in characters
   * y:     ypos in pixels, if >= 10000 in characters
   * fg:    foreground color
   * bg:    background color
   */
  int line,pixel;

  if((charn < ' ') || (charn > '~')) charn=' ';

  for(line = 0; line < 16; line++)
  {
    for(pixel = 0; pixel < 8; pixel++)
    {
      if((1 << (7-pixel) & charset8x16[(charn-32)*16+line]))
      {
        vram_s[(y+line)*DC_screen_size_X+(x+pixel)]=fg;
      }
      else
      {
        vram_s[(y+line)*DC_screen_size_X+(x+pixel)]=bg;
      }
    }
  }
}

void printxy(const char *text,int x, int y, uint16 fg, uint16 bg)
{
  /*
   * print some text on screen
   * if x or y position is > 10000 placement is done in characters, not pixels
   * if string starts with '\n' 8x16 charset is used instead of 8x8 one
   * 
   * text: text to print
   * x: x position
   * y: y position
   * fg: text color
   * bg: background color
   */
  
  const char *s;
  int xp,yp;
  
  if(x >= 10000) x=(x % 10000) * 8;

  if(*text == '\n')
  {
    if(y >= 10000) y=(y % 10000) * 16;
    for(s=text+1;*s;s++)
    {
      draw_char_d16((int)*s,x,y,fg,bg);
      x+=8;
      if(x > DC_screen_size_X-8) break;
    }
  }
  else
  {
    if(y >= 10000) y=(y % 10000) * 8;
    for(s=text;*s;s++)
    {
      draw_char_s16((int)*s,x,y,fg,bg);
      x+=8;
      if(x > DC_screen_size_X-8) break;
    }
  }
}

void printint(int value,int x, int y, uint16 fg, uint16 bg)
{
  /*
   * hack for easy printing an integer in decimal
   */
  char tmpbuffer[11];
  sprintf(tmpbuffer,"%d",value);
  printxy(tmpbuffer,x,y,fg,bg);
}

void printhex(uint32 value,int x, int y, uint16 fg, uint16 bg)
{
  /*
   * hack for easy printing an integer in hexadecimal
   */
  char tmpbuffer[9];
  sprintf(tmpbuffer,"%08x",value);
  printxy(tmpbuffer,x,y,fg,bg);
}

void error_flash(int color)
{
  /*
   * quick and dirty error (border) flashing routine
   * 
   * color: 1=blue, 2=green, 4=red, etc.
   */
  
  int r=0,g=0,b=0,i,l;
  if(color & 4) r=1;
  if(color & 2) g=1;
  if(color & 1) b=1;
  for(l=0;l<2;l++)
  {
    for(i=0;i<256;i++)
    {
      vid_waitvbl();
      vid_border_color(i*r,i*g,i*b);
    }
    for(i=254;i>0;i--)
    {
      vid_waitvbl();
      vid_border_color(i*r,i*g,i*b);
    }
  }
}


int display_targa(const void *targa,int xpos,int ypos,int pixel_mode)
{
  /*
   * load targa file and draw it on screen
   * if pos > 0 placement from top/left; (1,1) is top/left
   * if pos < 0 placement from bottom/right (-1,-1) is bottom/right
   * if pos = 0 placement is centered
   * 
   * filename: name of targa file
   * xpos: x position of picture
   * ypos: y position of picture
   * pixel_mode: PM_RGB555, PM_RGB565 or PM_RGB888
   */
  
  uint8 *image;
  uint8 image_data_offset,topleft,interlace;
  uint16 size_x,size_y,x,y,row;
  uint32 screen_offset;
  uint32 fd;
  uint32 image_data_pos,screen_data_pos;
  image = (uint8*)targa;
  
  image_data_offset = 18+image[0]; // Skip ID-Field
  if(image[1] != 0x00) return 1;
  if(image[2] != 0x02) return 1;
  size_x = image[12] + ( image[13] << 8 );
  size_y = image[14] + ( image[15] << 8 );
  if(size_x > 640)     return 2;
  if(size_y > 480)     return 2;
  if(image[16] != 24)  return 2;
  
  if(xpos == 0) 
  {
    x=(DC_screen_size_X-size_x)/2;
  }
  else
  {
    if(xpos > 0)
    {
      x=xpos-1;
    }
    else
    {
      x=DC_screen_size_X-size_x+1;
    }
  }
  
  if(ypos == 0) 
  {
    y=(DC_screen_size_Y-size_y)/2;
  }
  else
  {
    if(ypos > 0)
    {
      y=ypos-1;
    }
    else
    {
      y=DC_screen_size_Y-size_y+1;
    }
  }
  
  screen_offset=x+y*DC_screen_size_X;

  topleft = (image[17] & 0x20) >> 5;
  interlace = (image[17] & 0xc0) >> 6;

  for (y=0; y< size_y; y++) {
    switch(interlace)
    {
     case 2:  // four pass interlace
      if      (size_y < (1*size_y) / 4) row = 4 *  y;
      else if (size_y < (2*size_y) / 4) row = 4 * (y - ((1*size_y)/4)) + 1;
      else if (size_y < (3*size_y) / 4) row = 4 * (y - ((2*size_y)/4)) + 2;
      else                              row = 4 * (y - ((3*size_y)/4)) + 3;
      break;
     case 1:  // two pass interlace 
      if      (y < size_y / 2) row = 2 *  y;
      else                     row = 2 * (y - size_y/2) + 1;
      break;
     default: // no interlace
      row = y;
    }
    
    image_data_pos = image_data_offset+(y*size_x*3);
    if (topleft)
    {
      screen_data_pos = screen_offset+row*DC_screen_size_X;
    }
    else
    {
      /* bottom-left origin: invert y */
      screen_data_pos = screen_offset+(size_y - row - 1)*DC_screen_size_X;
    }
    
    for(x=0;x<size_x;x++)
    {
      switch(pixel_mode)
      {
       case PM_RGB555:
	vram_s[screen_data_pos++]=
	  ((image[image_data_pos+0] & 0xf8 ) >> 3) | // B
	  ((image[image_data_pos+1] & 0xf8 ) << 2) | // G
	  ((image[image_data_pos+2] & 0xf8 ) << 7) ; // R
	break;
       case PM_RGB565:
	vram_s[screen_data_pos++]=
	  ((image[image_data_pos+0] & 0xf8 ) >> 3) | // B
	  ((image[image_data_pos+1] & 0xfc ) << 3) | // G
	  ((image[image_data_pos+2] & 0xf8 ) << 8) ; // R
	break;
       case PM_RGB888:
	vram_l[screen_data_pos++]=
	  (image[image_data_pos+0]      ) | // B
	  (image[image_data_pos+1] <<  8) | // G
	  (image[image_data_pos+2] << 16) ; // R
	break;
      }
      image_data_pos += 3;
    }
  }
  return 0;
}


uint32 select_one_from_list(char **list, int cont, uint32 buttonmask,
			    int *actualline, int *topoffset,
			    int x,int y, int width, int height,
			    uint16 txtfg, uint16 txtbg,
			    uint16 barfg, uint16 barbg)
{
  /*
   * Present a list with the option to choose one
   * list: the list of options
   * x:    xpos
   * y:    ypos
   * width: in characters
   * height: in characters (neg for 8pixel font)
   * actualline: pre-selection (CBR->RETURN)
   * topoffset: where to start display the list (CBR->RETURN)
   * txtfg: foreground color
   * txtbg: background color
   * barfg: selected foreground color
   * barbg: selected background color
   */
  int line,i,endoflist,buttondelay,boundry=0,charheight;
  uint32 buttons;

  if(height >=  3) boundry=1;
  if(height >=  7) boundry=2;
  if(height >= 11) boundry=3;
  
  char tmpl[16],printbuf[256];
  if(height >= 0)
  {
    sprintf(tmpl,"\n %%-%02ds ",width);
    charheight=16;
  }
  else
  {
    height=-height;
    sprintf(tmpl," %%-%02ds ",width);
    charheight=8;
  }
  
 
  for(endoflist=0;list[endoflist];endoflist++);
  
  for(;;)
  {
    if((*actualline) >= endoflist)
      (*actualline)-=endoflist;
    if((*actualline) < 0)
      (*actualline)+=endoflist;
    
    if((*actualline) >= (*topoffset)+height)
      (*topoffset) = (*actualline)-height+1;
    if((*actualline) < (*topoffset))
      (*topoffset) = (*actualline);
    
    if((*actualline) >= (*topoffset)+height-boundry-1)
      (*topoffset) = (*actualline)-height+boundry+1;
    if((*actualline) < (*topoffset)+boundry)
      (*topoffset) = (*actualline)-boundry;
    
    if((*topoffset) >= endoflist-height)
      (*topoffset) = endoflist-height;
    if((*topoffset) < 0)
      (*topoffset) = 0;

    gdrom_check_lid();
    vid_waitvbl();
    
    for(line=0;line<height;line++)
    {
#if 0
      if((line+(*topoffset)) < endoflist)
      {
	sprintf(printbuf,tmpl,list[line+*topoffset]);
      }
      else
      {
	sprintf(printbuf,tmpl,"");
      }
#else
      sprintf(printbuf,tmpl,list[line+*topoffset]);
      if(strlen(printbuf) > width)
      {
	printbuf[width-2]   = '\0';
//	printbuf[width-3] = ' ';
      }
      
#endif
      
      if((line+(*topoffset)) == (*actualline))
      {
	printxy(printbuf,x,y+line*charheight,barfg,barbg);
      }
      else
      {
	printxy(printbuf,x,y+line*charheight,txtfg,txtbg);
      }
      
    }
    buttons=get_cont_buttons(cont);
    
    if((buttondelay-- < 0) || !buttons)
      buttondelay=2;

   
    if(buttons & buttonmask)
    {
      return buttons;
    }
    
    if(buttondelay == 1)
    {
      if(buttons & CONT_DPAD_DOWN)
      {
	(*actualline)++;
      }
      
      if(buttons & CONT_DPAD_UP)
      {
	(*actualline)--;
      }
      
      if(buttons & CONT_DPAD_RIGHT)
      {
	(*actualline)+=height;
	(*topoffset)+=height;
      }
      
      if(buttons & CONT_DPAD_LEFT)
      {
	(*actualline)-=height;
	(*topoffset)-=height;
      }
    }
  }
}

uint32 get_cont_buttons(int cont)
{
  /*
   * get the buttons of a controller for menu, etc.
   * 
   * cont: controller number (0..3)
   * 
   * returns bitfield with 1=pressed
   */
  
  static uint8 mcont[4] = {0,0,0,0};
  cont_cond_t cond;

  if (!mcont[cont])
  {
    mcont[cont] = maple_create_addr(0,0);
    if (!mcont[cont])
    {
//      printf("No controllers attached\n");
      return 0x10000;
    }
  }

  /* Check for start on the controller */
  if (cont_get_cond(mcont[cont], &cond))
  {
//    printf("Error getting controller status\n");
    return 0x20000;
  }

  return (~cond.buttons) & 0xffff;
}

