/*
 * This file is part of the DCStella project
 * location: dcstella/core/sndDC.cxx
 * written by SvOlli, based on original Stella SndUnix.cxx
 */

#include "sndDC.hxx"

extern "C"
{
#include "o/armcode.h"
#include <dream.h>
#include "screen.h" // for debugging
#include "o/compilemode.h"
}

#define AICA_OFFSET(x) (0xa0800000+x)


SoundDC::SoundDC()
     :myMute(false),
      cmdp((unsigned long*)AICA_OFFSET(0xf000))
{
  // Initialize to impossible values so they will be reset 
  // on the first call to the set method
  myAUDC0 = 255;
  myAUDC1 = 255;
  myAUDF0 = 255;
  myAUDF1 = 255;
  myAUDV0 = 255;
  myAUDV1 = 255;

  snd_load_arm(armcode, sizeof(armcode));
}

 
SoundDC::~SoundDC()
{
  snd_stop_arm();
}


void SoundDC::set(Sound::Register reg, uInt8 value)
{
  unsigned long command;

  switch(reg)
  {
   case AUDC0:
    {
      if(myAUDC0 != (value & 0x0f))
      {
        myAUDC0 = (value & 0x0f);
        command = myAUDC0 | 0x00;
      }
      else
      {
        return;
      }
      break;
    }
    
   case AUDC1:
    {
      if(myAUDC1 != (value & 0x0f))
      {
        myAUDC1 = (value & 0x0f);
        command = myAUDC1 | 0x20;
      }
      else
      {
        return;
      }
      break;
    }
    
   case AUDF0:
    {
      if(myAUDF0 != (value & 0x1f))
      {
        myAUDF0 = (value & 0x1f);
        command = myAUDF0 | 0x40;
      }
      else
      {
        return;
      }
      break;
    }
    
   case AUDF1:
    {
      if(myAUDF1 != (value & 0x1f))
      {
        myAUDF1 = (value & 0x1f);
        command = myAUDF1 | 0x60;
      }
      else
      {
        return;
      }
      break;
    }
    
   case AUDV0:
    {
      if(myAUDV0 != (value & 0x0f))
      {
        myAUDV0 = (value & 0x0f);
        command = myAUDV0 | 0x80;
      }
      else
      {
        return;
      }
      break;
    }
    
   case AUDV1:
    {
      if(myAUDV1 != (value & 0x0f))
      {
        myAUDV1 = (value & 0x0f);
        command = myAUDV1 | 0xA0;
      }
      else
      {
        return;
      }
      break;
    }
    
   default:
    break;
  }

  if(!myMute)
  {
    snd_send_command(0x01000000 | command);
  }
}


void SoundDC::mute(bool state)
{
  myMute = state;
  snd_send_command((state) ? 0x010000e1 : 0x010000e0);
}


void SoundDC::snd_send_command(unsigned long cmd)
{
  while(*(SoundDC::cmdp));
  *SoundDC::cmdp = cmd;
  SoundDC::cmdp++;
  if(SoundDC::cmdp >= (unsigned long*)AICA_OFFSET(0xffff))
  {
    SoundDC::cmdp = (unsigned long*)AICA_OFFSET(0xf000);
  }
}
