/*
 * This file is part of the DCStella project
 * location: dcstella/core/dcload-console.c
 * written by SvOlli, based on original Stella sources SndUnix.hxx
 */

#ifndef SOUNDDC_HXX
#define SOUNDDC_HXX

#include "bspf.hxx"
#include "Sound.hxx"

class SoundDC : public Sound
{
  public:
    /**
      Create a new sound object
    */
    SoundDC();
 
    /**
      Destructor
    */
    virtual ~SoundDC();

  public: 
    /**
      Set the value of the specified sound register

      @param reg The sound register to set
      @param val The new value for the sound register
    */
    virtual void set(Sound::Register reg, uInt8 val);
  
    /**
      Set the mute state of the sound object

      @param state Mutes sound iff true
    */
    virtual void mute(bool state);

    volatile unsigned long *cmdp;

  private:
  
    void snd_send_command(unsigned long cmd);
  
    // Indicates if the sound is muted or not
    bool myMute;

  
    // Buffers for the audio registers used so we only 
    // send "changes" to the stella-sound process
    uInt8 myAUDC0;
    uInt8 myAUDC1;
    uInt8 myAUDF0;
    uInt8 myAUDF1;
    uInt8 myAUDV0;
    uInt8 myAUDV1;
};
#endif

